"""
Imperfect example solution for the afternoon Exercise of Day 1.

This module contains the elements for an abstract graph representation.

It is not intended to be a good design, so go and improve it :-)
"""

import heapq


# Question: Does this class make sense, do we really need it?
class Node(object):
    """Base class for nodes.

    To add additional data attributes to nodes you can derive subclasses.
    """

    def __init__(self):
        self._out_edges = []  # outgoing edges

    @property
    def out_edges(self):
        return self._out_edges


# Question: How does this class work together with the Node class?
#    Are there negative aspects of this design?
class Edge(object):
    """Class representing a directed edge connection between two nodes."""

    def __init__(self, head, tail):
        """
        head, tail -- The end and starting node of this edge, or can be None.
        """
        self._head = head
        self._tail = tail
        tail._out_edges.append(self)

    @property
    def head(self):
        return self._head

    @property
    def tail(self):
        return self._tail


class NoPathGraphException(Exception):
    """Exception signaling that there is no path between nodes."""
    pass


class Graph(object):
    """Class to represent a complete graph with nodes and edges.

    Note that this class does not support nodes which do not belong to any
    edge.
    """

    def __init__(self, edges):
        self.edges = edges

    @property
    def nodes(self):
        """Return set of nodes in this graph."""
        nodes = set()
        for edge in self.edges:
            nodes.add(edge.head)
            nodes.add(edge.tail)
        return nodes

    # Note: This method uses the so called "Template Method" pattern.
    #    This means that the part that varies (the weight_func) has been
    #    extracted into a separate method. In order to customize it one has
    #    to derive a new subclass where the method is overriden with a
    #    custom weight_func.
    # Question: Is this a good design? Should this be part of the Graph class?
    # Note: The values of the shortest_paths dictionary are an example where
    #    we use tuples instead of a special class.
    @classmethod  # If you don't know this decorator yet don't worry about it.
    def shortest_path(cls, start_node, end_node):
        """Return the shortest path and its overall weight.

        start_node -- Starting node of the path.
        end_node -- End node of the path.
        """
        ## this is basically Dijkstra's algorithm
        # store the shortest path to all nodes,
        # the value tuples contain the path length and the previous node
        shortest_paths = {start_node: (0, None)}
        # we use this list as a priority cue with heapq
        # the tuples contain the overall path length and the edge itself
        edge_heap = [(cls.weight_func(edge), edge)
                     for edge in start_node.out_edges]
        heapq.heapify(edge_heap)
        while edge_heap:
            path_weight, edge = heapq.heappop(edge_heap)
            if (edge.head not in shortest_paths or
                shortest_paths[edge.head][0] > path_weight):
                shortest_paths[edge.head] = (path_weight, edge)
                # if we already visited this node then there may be edge
                # duplicates in the heap, but this is no problem because
                # the newer edges are guaranteed to come first
                for out_edge in edge.head.out_edges:
                    heapq.heappush(
                        edge_heap,
                        (path_weight + cls.weight_func(out_edge), out_edge))
        if end_node not in shortest_paths:
            msg = ("The is no connection from node {} to node {}".
                   format(start_node, end_node))
            raise NoPathGraphException(msg)
        # assemble the shortest path from end to start
        path_weight = shortest_paths[end_node][0]
        path_edges = [shortest_paths[end_node][1]]
        current_node = path_edges[-1].tail
        while current_node is not start_node:
            path_edges.append(shortest_paths[current_node][1])
            current_node = path_edges[-1].tail
        return path_edges[::-1], path_weight

    @staticmethod  # If you don't know this decorator yet don't worry about it.
    def weight_func(edge):
        """Return the weight number for a given edge.

        Override this method in subclasses to define custom weight functions.
        """
        return 1
