"""
Unittests for the graph module.
"""

import unittest
from graph import Node, Edge, Graph, NoPathGraphException


class TestEdge(unittest.TestCase):

    def test_edge_creation(self):
        """Test the basic edge creation."""
        node1 = Node()
        node2 = Node()
        edge1 = Edge(node2, node1)
        edge2 = Edge(node1, node2)
        self.assertEqual(set(node1.out_edges), set([edge1]))
        edge3 = Edge(node2, node1)
        self.assertEqual(set(node1.out_edges), set([edge1, edge3]))


class TestGraph(unittest.TestCase):

    def test_graph_creation(self):
        """Test the basic construction of a graph."""
        node1 = Node()
        node2 = Node()
        edge1 = Edge(node2, node1)
        edge2 = Edge(node1, node2)
        graph = Graph([edge1, edge2])
        self.assertEqual(set(graph.nodes), set([node2, node1]))

    def test_shortest_path(self):
        """Test the shortest path algorithm for a simple example."""
        nodes = [Node() for _ in range(5)]
        # create a diamond shape
        edges = [
            Edge(nodes[1], nodes[0]),
            Edge(nodes[2], nodes[1]),
            Edge(nodes[4], nodes[2]),
            Edge(nodes[3], nodes[0]),
            Edge(nodes[4], nodes[3]),
        ]
        graph = Graph(edges)
        path, length = graph.shortest_path(nodes[0], nodes[4])
        self.assert_(path == [edges[3], edges[4]])
        self.assert_(length == 2)
        self.assertRaises(NoPathGraphException,
                          lambda: graph.shortest_path(nodes[4], nodes[0]))

    # TODO: add more tests for the shortest path algorithm,
    #    especially with respect to path weight updates
